import copy
import dataclasses
import logging
import typing
from .cache_manager import CachedOutputs
from .variable import Variable

logger = logging.getLogger(__name__)


class Context:
    """Manage variables for a single experiment."""
    def __init__(self, environment_variables: typing.Dict[str, str] = None, cache_manager=None):
        self._envs = copy.deepcopy(environment_variables or {})
        self._cache_manager = cache_manager
        self._outputs = {}

    def add_outputs(self, name: str, outputs: typing.Union[dataclasses.dataclass, CachedOutputs]):
        """Add Task outputs to the context so that subsequent Tasks can consume them.

        Args:
            name (str): the name of the task.
            outputs (Outputs or CachedOutputs instance): the outputs of the task.
        """
        if name in self._outputs:
            logger.warning(f"Duplicated task name: {name}. The outputs are overwritten.")
        self._outputs[name] = outputs

    def get_outputs(self, output_name: str):
        """Get the outputs of previous tasks.
        Args:
            name (str): the name of the task
        Returns:
            Outputs dataclass. If the task had been skipped, returns CachedOutputs.
        """
        if output_name not in self._outputs:
            raise ValueError(f"Output {output_name} is not found.")
        return self._outputs[output_name]

    def add_environment_variable(self, name: str, value: str):
        self._envs[name] = value

    def get_environment_variable(self, name: str):
        if name not in self._envs:
            raise ValueError(f"Environment variable {name} is not found.")
        return self._envs[name]

    def resolve(self, value):
        """Recursively replace Variables with actual value.

        Supported containers are:
            - Dict
            - List
            - dataclasses.dataclass
        """
        if isinstance(value, dict):
            return {k: self.resolve(v) for k, v in value.items()}
        elif dataclasses.is_dataclass(value):
            return type(value)(**{field.name: self.resolve(getattr(value, field.name)) for field in dataclasses.fields(value)})
        elif isinstance(value, list):
            return [self.resolve(v) for v in value]
        elif isinstance(value, Variable):
            return value.resolve(self)
        else:
            return value

    def get_cached_outputs(self, task_name, task_version, task_hash: str, outputs_class: dataclasses.dataclass) -> typing.Optional[CachedOutputs]:
        """Try to get cached outputs for the given task

        Args:
            task_name (str): Task name
            task_version (str): Task version
            config (Task.Config): Task Config dataclass instance. It can contain Variables.
            inputs (Task.Inputs): Task Inputs dataclass instance. It can contain Variables.

        Returns:
            CachedOutput instance if there is a cache. If not, returns None.
        """
        if not self._cache_manager:
            return None

        logger.debug(f"Trying to get cache for Task {task_name} version {task_version}. Hash: {task_hash}")
        return self._cache_manager.get_cache(task_name, task_version, task_hash, outputs_class)

    def add_cache_outputs(self, task_name, task_version, task_hash: str, outputs):
        """Save the task outputs to the cache storage."""
        if self._cache_manager:
            logger.debug(f"Uploading cache for Task {task_name} version {task_version}. Hash: {task_hash}")
            self._cache_manager.upload_cache(task_name, task_version, task_hash, outputs)

    def clone(self):
        return copy.deepcopy(self)



import dataclasses
from typing import Any, Dict, List, Optional
import typing


@dataclasses.dataclass
class TaskDescription:
    task: str
    name: Optional[str] = None  # Optional name for the task. Must be unique in the job. This name will be used for OutputVariable names.
    inputs: Optional[Dict[str, Any]] = None
    config: Optional[Dict[str, Any]] = None

    @classmethod
    def from_dict(cls, data: typing.Dict):
        return cls(**data)


@dataclasses.dataclass
class JobDescription:
    tasks: List[TaskDescription]
    on_error: Optional[List[TaskDescription]]

    @classmethod
    def from_dict(cls, data: typing.Dict):
        c = {'tasks': [TaskDescription.from_dict(t) for t in data['tasks']],
             'on_error': data.get('on_error', None) and [TaskDescription.from_dict(t) for t in data.get('on_error')]}
        return cls(**c)



import dataclasses
from irisml.core.context import Context


class TaskBase:
    """Base class for Task definition.

    Attributes:
        VERSION (str): Task must overwrite this atttribute. The format is "(major).(minor).(patch)".
        CACHE_ENABLED (bool): Set False if Task cannot be cached.
        RESOLVE_CONFIG_VARIABLES (bool): If set False, variables in the task config will not be replaced by actual values.
    """
    VERSION = '0.0.0'
    CACHE_ENABLED = True
    RESOLVE_CONFIG_VARIABLES = True

    @dataclasses.dataclass(frozen=True)
    class Config:
        """Configuration for the task. Must be overwritten if the task requires a config"""
        pass

    @dataclasses.dataclass(frozen=True)
    class Inputs:
        """Inputs type. Must be overwritten if the task requires inputs. Nested dataclass is not allowed."""
        pass

    @dataclasses.dataclass(frozen=True)
    class Outputs:
        """Outputs type. Must be overwritten if the task has outputs. Nested dataclass is not allowed."""
        pass

    def __init__(self, config: Config, context: Context = None):
        self._config = config
        self._context = context or Context()

    @property
    def config(self):
        return self._config

    @property
    def context(self):
        return self._context

    def execute(self, inputs: Inputs) -> Outputs:
        """This method must be overwritten by the task."""
        raise NotImplementedError

    def dry_run(self, inputs: Inputs) -> Outputs:
        """This method will be called in dry-run mode. Task can overwrite this method.

        This method must return immediately.
        """
        try:
            return self.Outputs()
        except TypeError as e:
            raise RuntimeError("Failed to instantiate an Outputs object. You must set default value for all fields or define dry_run() method.") from e



import contextlib
import dataclasses
import io
import sys
import unittest
import unittest.mock
import irisml.core
from irisml.core.job_runner import JobRunner


class FakeTask(irisml.core.TaskBase):
    def execute(self, inputs):
        raise FakeException('fake_exception')


class FakeOnErrorTask(irisml.core.TaskBase):
    @dataclasses.dataclass
    class Inputs:
        exception: Exception

    def execute(self, inputs):
        print("The error is handled.")
        return self.Outputs()


class FakeException(RuntimeError):
    pass


class TestJobRunner(unittest.TestCase):
    def test_error_handler_is_called(self):
        job_dict = {'tasks': [{'task': 'fake_task'}], 'on_error': [{'task': 'error_handler', 'inputs': {'exception': '$env.IRISML_EXCEPTION'}}]}
        runner = JobRunner(job_dict, {})
        with unittest.mock.patch.dict(sys.modules, {'irisml.tasks.fake_task': unittest.mock.MagicMock(Task=FakeTask), 'irisml.tasks.error_handler': unittest.mock.MagicMock(Task=FakeOnErrorTask)}):
            with contextlib.redirect_stdout(io.StringIO()) as f:
                with self.assertRaises(FakeException):
                    runner.run()
            self.assertEqual(f.getvalue().strip(), 'The error is handled.')



import dataclasses
import importlib
import logging
import random
import time
import typing
import torch
from irisml.core import TaskDescription
from .hash_generator import HashGenerator
from .task_base import TaskBase
from .variable import replace_variables, Variable


logger = logging.getLogger(__name__)


class Task:
    """Represents a task. It doesn't require actual task modules until load_module() is called."""
    def __init__(self, description: TaskDescription):
        assert description.task.islower()

        self._task_name = description.task
        self._inputs_dict = replace_variables(description.inputs or {})
        self._config_dict = replace_variables(description.config or {})
        self._name = description.name or self._task_name
        self._task_class = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def task_name(self):
        return self._task_name

    def execute(self, context, dry_run=False):
        if dry_run:
            return self.dry_run(context)

        if not self._task_class:
            raise RuntimeError("load_module() must be called before executing the task.")

        config = self._load_config(self._task_class.Config, self._config_dict)
        inputs = self._load_inputs(self._task_class.Inputs, self._inputs_dict)

        if self._task_class.CACHE_ENABLED:
            task_hash = HashGenerator.calculate_hash([config, inputs], context)
            cached_outputs = context.get_cached_outputs(self._task_name, self._task_class.VERSION, task_hash, self._task_class.Outputs)
            if cached_outputs:
                logger.info(f"[{self._task_name}]: Found cached outputs. Skipping the task.")
                context.add_outputs(self.name, cached_outputs)
                return cached_outputs

        logger.info(f"[{self._task_name}] Started.")
        start_time = time.time()
        # If RESOLVE_CONFIG_VARIABLES is False, the task will be responsible for resolving the configs.
        assert self._task_class.RESOLVE_CONFIG_VARIABLES or not self._task_class.CACHE_ENABLED
        resolved_config = context.resolve(config) if self._task_class.RESOLVE_CONFIG_VARIABLES else config
        resolved_inputs = context.resolve(inputs)

        if self._task_class.CACHE_ENABLED:
            if HashGenerator.calculate_hash(config, context) != HashGenerator.calculate_hash(resolved_config):
                logger.error("Resolved config has different hash.")
            if HashGenerator.calculate_hash(inputs, context) != HashGenerator.calculate_hash(resolved_inputs):
                logger.error("Resolved inputs has different hash.")

        self._reset_random_seed()
        logger.debug(f"Instantiating the task module. config={resolved_config}")
        task = self._task_class(resolved_config, context)
        outputs = task.execute(resolved_inputs)
        if outputs is None:
            raise RuntimeError(f"The task {self._task_name} returned None.")
        if not isinstance(outputs, self._task_class.Outputs):
            raise RuntimeError(f"Task {self._task_name} returned invalid outputs: {outputs}")

        logger.info(f"[{self._task_name}] Completed. {time.time() - start_time:.3f}s")

        context.add_outputs(self.name, outputs)
        if self._task_class.CACHE_ENABLED:
            context.add_cache_outputs(self._task_name, self._task_class.VERSION, task_hash, outputs)

        return outputs

    def dry_run(self, context):
        """Dry run the task. Task can define its own dry_run() method."""
        if not self._task_class:
            raise RuntimeError("load_module() must be called before executing the task.")

        config = self._load_config(self._task_class.Config, self._config_dict)
        inputs = self._load_inputs(self._task_class.Inputs, self._inputs_dict)

        assert self._task_class.RESOLVE_CONFIG_VARIABLES or not self._task_class.CACHE_ENABLED
        resolved_config = context.resolve(config) if self._task_class.RESOLVE_CONFIG_VARIABLES else config
        resolved_inputs = context.resolve(inputs)

        logger.debug(f"Instantiating the task module. config={resolved_config}")
        task = self._task_class(resolved_config, context)
        outputs = task.dry_run(resolved_inputs)

        if not isinstance(outputs, self._task_class.Outputs):
            raise RuntimeError(f"Task {self._task_name} returned invalid outputs: {outputs}")

        context.add_outputs(self.name, outputs)
        return outputs

    def load_module(self):
        """Load a task module dynamically. If the module was not found, throws a RuntimeError"""
        task_module_name = 'irisml.tasks.' + self.task_name
        try:
            task_module = importlib.import_module(task_module_name)
        except ModuleNotFoundError as e:
            if e.name == task_module_name:
                raise RuntimeError(f"Task not found: {task_module_name}") from e
            else:
                raise

        task_class = getattr(task_module, 'Task')
        if not issubclass(task_class, TaskBase):
            raise RuntimeError(f"Failed to load {self.task_name}. Please make sure the Task class inherits the TaskBase class.")
        self._task_class = task_class

        # Verify that the config is loadable.
        self._load_config(task_class.Config, self._config_dict)
        self.validate()

    def validate(self):
        """Check if the task satisfies the rules."""
        if not self._task_class:
            raise RuntimeError("load_module() must be called first.")

        if self._task_class.VERSION == '0.0.0':
            logger.warning(f"Task {self._task_name} is version 0.0.0. Please define VERSION attribute in the Task class..")

        if not dataclasses.is_dataclass(self._task_class.Config):
            raise RuntimeError(f"Config class must be a dataclass. Actual: {type(self._task_class.Config)}")
        if not dataclasses.is_dataclass(self._task_class.Inputs):
            raise RuntimeError(f"Inputs class must be a dataclass. Actual: {type(self._task_class.Inputs)}")
        if not dataclasses.is_dataclass(self._task_class.Outputs):
            raise RuntimeError(f"Outputs class must be a dataclass. Actual: {type(self._task_class.Outputs)}")

        for f in dataclasses.fields(self._task_class.Inputs):
            if dataclasses.is_dataclass(f.type):
                raise RuntimeError(f"Nested input dataclass is not allowed: {f.name}")

        for f in dataclasses.fields(self._task_class.Outputs):
            if dataclasses.is_dataclass(f.type):
                raise RuntimeError(f"Nested output dataclass is not allowed: {f.name}")

    @staticmethod
    def _load_config(config_class: dataclasses.dataclass, config_dict):  # noqa: C901 # TODO
        """Initialize a config class loaded from the actual task module."""
        assert dataclasses.is_dataclass(config_class)

        def load(type_class, value):
            origin = typing.get_origin(type_class)
            args = typing.get_args(type_class)
            if dataclasses.is_dataclass(type_class):
                assert isinstance(value, dict)
                c = {}
                for field in dataclasses.fields(type_class):
                    if field.name in value:
                        c[field.name] = load(field.type, value[field.name])

                unused_keys = set(value.keys()) - set(c.keys())
                if unused_keys:
                    raise ValueError(f"There are redundant fields: {unused_keys}")
                return type_class(**c)
            elif origin is tuple:
                return tuple(load(arg, v) for arg, v in zip(args, value))
            elif origin is list:
                if isinstance(value, Variable):
                    value.expected_type = list
                    return value
                return [load(args[0], v) for v in value]
            elif origin is dict:
                if isinstance(value, Variable):
                    value.expected_type = dict
                    return value
                return {load(args[0], k): load(args[1], v) for k, v in value.items()}
            elif origin is typing.Union:  # Optional[X] is Union[X, NoneType]
                if len(args) == 2 and type(None) in args:
                    expected_type = next(a for a in args if a is not type(None))  # noqa: E721
                    return load(expected_type, value)
                elif type(value) in args:
                    return value
                else:
                    raise ValueError(f"Union type is not allowed: {type_class}")
            elif type_class == typing.Any:
                return value
            elif origin is None:
                if isinstance(value, Variable):
                    value.expected_type = type_class
                    return value
                return type_class(value)
            elif origin is typing.Literal:
                if isinstance(value, Variable):
                    value.expected_type = type_class
                elif value not in args:
                    raise ValueError(f"Unexpected literal value: {value}. Expected: {args}")
                return value
            else:
                raise ValueError(f"Config data type is not supported: {type_class} value: {value}")

        return load(config_class, config_dict)

    @staticmethod
    def _load_inputs(inputs_class: dataclasses.dataclass, inputs_dict: dict):
        c = {}
        # Inputs doesn't have nested values.
        for field in dataclasses.fields(inputs_class):
            if field.name in inputs_dict:
                if isinstance(inputs_dict[field.name], Variable):
                    inputs_dict[field.name].expected_type = field.type
                c[field.name] = inputs_dict[field.name]
        return inputs_class(**c)

    def _reset_random_seed(self):
        """Reset the random seed.

        To make sure the tasks are deterministic, we reset the random seed every time a task starts.
        """
        torch.manual_seed(42)
        random.seed(42)

    def __str__(self):
        return f"Task {self.task_name} (name: {self.name})"



import argparse
import json
import pathlib
import pickle
from irisml.core import Context, Task, TaskDescription
from irisml.core.commands.common import configure_logger


class KeyValuePairAction(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        assert len(values) == 1
        k, v = values[0].split('=', 2)
        dest = getattr(args, self.dest, None)
        if not dest:
            dest = {}
            setattr(args, self.dest, dest)

        dest[k] = v


def main():
    """TODO This method is WIP."""
    configure_logger()
    parser = argparse.ArgumentParser(description="Run a single task")
    parser.add_argument('task_name')
    parser.add_argument('--inputs', help="JSON-encoded inputs.")
    parser.add_argument('--inputs_pickle', nargs=1, action=KeyValuePairAction, metavar='KEY=VALUE', help="Add a serialized object to the inputs.")
    parser.add_argument('--inputs_bytes', nargs=1, action=KeyValuePairAction, metavar='KEY=VALUE', help="Add a bytes oject to the inputs.")
    parser.add_argument('--config', help="JSON-encoded task configs.")

    args = parser.parse_args()

    task_dict = {'task': args.task_name, 'inputs': {}}

    if args.inputs:
        task_dict['inputs'] = json.loads(args.inputs)

    if args.inputs_pickle:
        for key, value in args.inputs_pickle.items():
            data = pickle.loads(pathlib.Path(value).read_bytes())
            task_dict['inputs'][key] = data

    if args.inputs_bytes:
        for key, value in args.inputs_bytes.items():
            data = pathlib.Path(value).read_bytes()
            task_dict['inputs'][key] = data

    if args.config:
        task_dict['config'] = json.loads(args.config)

    task_description = TaskDescription.from_dict(task_dict)
    context = Context()
    task = Task(task_description)
    task.load_module()
    outputs = task.execute(context)
    print(outputs)
